@@ -8,10 +8,19 @@
 # This software may be used and distributed according to the terms of the
 # GNU General Public License version 2 or any later version.
 
-import os, error
-from i18n import _
-from node import short, hex
-import util
+from __future__ import absolute_import
+
+import collections
+import os
+
+from .i18n import _
+from .node import (
+    hex,
+    short,
+)
+from . import (
+    error,
+)
 
 def bisect(changelog, state):
     """find the next node (if any) for testing during a bisect search.
@@ -64,13 +73,13 @@ def bisect(changelog, state):
     if not ancestors: # now we're confused
         if (len(state['bad']) == 1 and len(state['good']) == 1 and
             state['bad'] != state['good']):
-            raise util.Abort(_("starting revisions are not directly related"))
-        raise util.Abort(_("inconsistent state, %s:%s is good and bad")
+            raise error.Abort(_("starting revisions are not directly related"))
+        raise error.Abort(_("inconsistent state, %s:%s is good and bad")
                          % (badrev, short(bad)))
 
     # build children dict
     children = {}
-    visit = util.deque([badrev])
+    visit = collections.deque([badrev])
     candidates = []
     while visit:
         rev = visit.popleft()
@@ -135,17 +144,17 @@ def bisect(changelog, state):
 def load_state(repo):
     state = {'current': [], 'good': [], 'bad': [], 'skip': []}
     if os.path.exists(repo.join("bisect.state")):
-        for l in repo.opener("bisect.state"):
+        for l in repo.vfs("bisect.state"):
             kind, node = l[:-1].split()
             node = repo.lookup(node)
             if kind not in state:
-                raise util.Abort(_("unknown bisect kind %s") % kind)
+                raise error.Abort(_("unknown bisect kind %s") % kind)
             state[kind].append(node)
     return state
 
 
 def save_state(repo, state):
-    f = repo.opener("bisect.state", "w", atomictemp=True)
+    f = repo.vfs("bisect.state", "w", atomictemp=True)
     wlock = repo.wlock()
     try:
         for kind in sorted(state):
