@@ -38,7 +38,7 @@ class httpsendfile(object):
         self.write = self._data.write
         self.length = os.fstat(self._data.fileno()).st_size
         self._pos = 0
-        self._total = self.length / 1024 * 2
+        self._total = self.length // 1024 * 2
 
     def read(self, *args, **kwargs):
         try:
@@ -51,7 +51,7 @@ class httpsendfile(object):
         # requires authentication. Since we can't know until we try
         # once whether authentication will be required, just lie to
         # the user and maybe the push succeeds suddenly at 50%.
-        self.ui.progress(_('sending'), self._pos / 1024,
+        self.ui.progress(_('sending'), self._pos // 1024,
                          unit=_('kb'), total=self._total)
         return ret
 
@@ -70,11 +70,7 @@ def readauthforuri(ui, uri, user):
         gdict[setting] = val
 
     # Find the best match
-    if '://' in uri:
-        scheme, hostpath = uri.split('://', 1)
-    else:
-        # py2.4.1 doesn't provide the full URI
-        scheme, hostpath = 'http', uri
+    scheme, hostpath = uri.split('://', 1)
     bestuser = None
     bestlen = 0
     bestauth = None
@@ -111,7 +107,9 @@ class HTTPResponse(httpclient.HTTPResponse):
 
 class HTTPConnection(httpclient.HTTPConnection):
     response_class = HTTPResponse
-    def request(self, method, uri, body=None, headers={}):
+    def request(self, method, uri, body=None, headers=None):
+        if headers is None:
+            headers = {}
         if isinstance(body, httpsendfile):
             body.seek(0)
         httpclient.HTTPConnection.request(self, method, uri, body=body,
@@ -130,6 +128,7 @@ class http2handler(urllib2.HTTPHandler, urllib2.HTTPSHandler):
         self.ui = ui
         self.pwmgr = pwmgr
         self._connections = {}
+        # developer config: ui.http2debuglevel
         loglevel = ui.config('ui', 'http2debuglevel', default=None)
         if loglevel and not _configuredlogging:
             _configuredlogging = True
@@ -215,7 +214,7 @@ class http2handler(urllib2.HTTPHandler, urllib2.HTTPSHandler):
                 path = '/' + path
             h.request(req.get_method(), path, req.data, headers)
             r = h.getresponse()
-        except socket.error, err: # XXX what error?
+        except socket.error as err: # XXX what error?
             raise urllib2.URLError(err)
 
         # Pick apart the HTTPResponse object to get the addinfourl
@@ -233,7 +232,11 @@ class http2handler(urllib2.HTTPHandler, urllib2.HTTPSHandler):
     def http_open(self, req):
         if req.get_full_url().startswith('https'):
             return self.https_open(req)
-        return self.do_open(HTTPConnection, req, False)
+        def makehttpcon(*args, **kwargs):
+            k2 = dict(kwargs)
+            k2['use_ssl'] = False
+            return HTTPConnection(*args, **k2)
+        return self.do_open(makehttpcon, req, False)
 
     def https_open(self, req):
         # req.get_full_url() does not contain credentials and we may
@@ -271,14 +274,13 @@ class http2handler(urllib2.HTTPHandler, urllib2.HTTPSHandler):
             if '[' in host:
                 host = host[1:-1]
 
-        if keyfile:
-            kwargs['keyfile'] = keyfile
-        if certfile:
-            kwargs['certfile'] = certfile
+        kwargs['keyfile'] = keyfile
+        kwargs['certfile'] = certfile
 
         kwargs.update(sslutil.sslkwargs(self.ui, host))
 
         con = HTTPConnection(host, port, use_ssl=True,
+                             ssl_wrap_socket=sslutil.wrapsocket,
                              ssl_validator=sslutil.validator(self.ui, host),
                              **kwargs)
         return con
